#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include<linux/mpage.h>


#include "myfs.h"


static int myfs_writepage(struct page *page, struct writeback_control *wbc);
static int myfs_readpage(struct file *file, struct page *page);
int myfs_write_begin(struct file *file, struct address_space *mapping,
loff_t pos, unsigned len, unsigned flags, struct page **pagep,
void **fsdata);


struct address_space_operations myfs_aops= {
	.readpage = myfs_readpage,
	.writepage = myfs_writepage,
	.write_begin = myfs_write_begin,
	.write_end = generic_write_end,
	 .sync_page = block_sync_page,

/*	.prepare_write = myfs_prepare_write,
	.commit_write = generic_commit_write,
	.bmap = myfs_bmap*/
};

sector_t myfs_bmap(struct address_space *mapping,sector_t block);
int myfs_get_block(struct inode *inode, sector_t iblock, struct buffer_head *bh_result, int create);


static inline void add_chain(Indirect *p, struct buffer_head *bh , __le32 *v)
{
	p->key = *(p->p = v);
	p->bh=bh;
}

static inline int verify_chain (Indirect *from , Indirect *to)
{
//	while(from <= to && from->key)

}






static inline Indirect *myfs_get_branch(struct inode *inode,int depth,int *offsets,
                                         Indirect chain[1],int *err)
{
	Indirect *p = chain;
	*err = 0 ;
	add_chain(chain,NULL,&(MYFS_INOI(inode)->blockno));
	//inode_info->blockno contains the disk block number --- 
	if(!p->key){
	//means no block number has been obtained
		printk(KERN_INFO"No disk Block number has been obtained in %s",__FUNCTION__);
		return p;
	}
	while(--depth)
	{
	/*
	 *In our case depth shoild be 1 and therefore in our case
	 *we should not reach here we have only one block number -- 
	* in inode_info->blockno itself -- 
	* we dont need to load any other block number 	 
	 * */
	printk(KERN_INFO"this is wrong place we have reached depth >1 %s ",__FUNCTION__);
	}
	return NULL;	
}





static int myfs_block_to_path(struct inode *inode,
                        sector_t block, int offsets[4], int *err)
{
	int n=0;
if(block <0 )
	printk(KERN_ALERT "MYFS block[%d] is less than zero in %s",(int)block,__FUNCTION__ );
else if (block >0)
	printk(KERN_ALERT"MYFS block[%d] is > than zero in %s",(int)block,__FUNCTION__ );
else{
	//Blosck number is zero -- Correct for our case 
	offsets[n++] = block;
	}
//depth = 1 is returned and offset[0] = 0 is returned 
return n;
}



int  __myfs_write_begin(struct file *file, struct address_space *mapping,
loff_t pos, unsigned len, unsigned flags, struct page **pagep,
void **fsdata)
{


/*block_write_begin takes care of the basic task of block allocation and
 * bringing partial write blocks uptodate first.
 *
 * If *pagep is not NULL, then block_write_begin uses the locked page
 * at *pagep rather than allocating its own. In this case, the page will
 * not be unlocked or deallocated on failure.
 */
 return  block_write_begin(file, mapping,pos,len,flags,pagep,fsdata,myfs_get_block);
}

int myfs_write_begin(struct file *file, struct address_space *mapping,
loff_t pos, unsigned len, unsigned flags, struct page **pagep,
void **fsdata)
{
	struct super_block *sb = NULL;
	struct inode * inode;	
	*pagep=NULL; //IMPORTANT
	if(file)
	{
		sb = file->f_dentry->d_inode->i_sb;
		inode = file->f_dentry->d_inode;
		
  printk(KERN_ALERT"MYFS[%s] filesize = %lld, pos = %lld, len = %u\n",__FUNCTION__,inode->i_size, pos, len);
		if((pos+len) <= sb->s_blocksize )
			return __myfs_write_begin(file, mapping, pos, len, flags, pagep, fsdata);
		else
		return -ENOSPC;
	}
	else
	{
		printk(KERN_ALERT"MYFS%s file object  does not exist ",__FUNCTION__);
		return -1;
	}
}


/*int myfs_prepare_write(struct file* file,struct page* page,unsigned from , unsigned to){
	struct super_block *sb = file->f_dentry->d_inode->i_sb;
	 if(!from){
         	if(to >= sb->s_blocksize)
	           return -ENOSPC;
      	}
      	else if (file->f_dentry->d_inode->i_size + (to - from) >= sb->s_blocksize )
                return -ENOSPC;
	
	return block_prepare_write(page,from,to,myfs_get_block);	

}*/



int myfs_get_block(struct inode *inode, sector_t iblock, struct buffer_head *bh_result, int create)
{
	Indirect chain[2];
	Indirect *partial;
	int offsets[4];
	int err = -EIO;


	int depth = myfs_block_to_path(inode,iblock,offsets,&err);
	if(!depth){
		return err;
	}
	partial = myfs_get_branch(inode,depth,offsets,chain,&err);	
	if(!partial){
		/*Simplest Case -- Block found, No allocation needed*/
		//if NULL has been returned from get_branch means right.			
		//chain[0]->key contains the disk block number 
		// but chain[0]->p-- address of MYFS_INOI(inode)->blockno
		// chain[0]->bh = NULL	
		/*Block Found !!!!*/
		map_bh(bh_result, inode->i_sb, chain[depth-1].key);
		//map_bh sets 3 things in bh --  
		//bh->b_bdev = sb->s_bdev;
		//bh->b_blocknr = block;
		// bh->b_size = sb->s_blocksize;
		
		partial = chain+depth-1;
		goto cleanup;
	}
	cleanup:
	while(partial>chain){
		brelse(partial->bh);
		partial--;		
	}	

	return err;
}

static int myfs_readpage(struct file *file, struct page *page)
{
return mpage_readpage(page, myfs_get_block);

}
static int myfs_writepage(struct page *page, struct writeback_control *wbc)
{
        return block_write_full_page(page, myfs_get_block, wbc);
}

sector_t myfs_bmap(struct address_space *mapping,sector_t block){
      return generic_block_bmap(mapping,block,myfs_get_block);
}

