#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>

#include "myfs.h"

int myfs_create (struct inode * parent_dir, struct dentry * dentry, int mode, struct nameidata *nd);
int myfs_permission(struct inode *inode, int mask);
int myfs_unlink(struct inode* parent_inode,struct dentry* dentry); 
struct dentry * myfs_lookup(struct inode* parent_inode,struct dentry * child_dentry,struct nameidata* nd );
extern struct inode * myfs_iget(struct super_block *sb,unsigned long ino);

extern struct file_operations myfs_file_ops;
extern struct file_operations myfs_dir_ops ;
extern struct address_space_operations myfs_aops;

extern int myfs_dir_add(struct inode* parent_dir,unsigned int ino,const char* name);
extern unsigned32 get_free_inode_number(struct myfs_super_block* ms);

struct inode_operations myfs_dir_inode_ops = {
				.create = myfs_create,
				.lookup = myfs_lookup,
			//	.permission = myfs_permission
				.unlink = myfs_unlink, 
};

struct inode_operations myfs_file_inode_ops = {
			.permission = myfs_permission
			};


extern int myfs_find_entry(struct inode* parent_dir, char* name);
int myfs_permission(struct inode *inode, int mask)
{
 return 0;
}

 int myfs_create (struct inode * parent_dir, struct dentry * dentry, int mode, struct nameidata *nd)
{
	struct super_block* sb = parent_dir->i_sb;
	struct myfs_sb_info*  msi =(struct myfs_sb_info*)sb->s_fs_info;
	struct myfs_super_block *ms = msi->ms;
	struct buffer_head* bh = msi->myfs_super_bh;
	struct inode * new_inode_p;
	struct myfs_inode_info * new_inoi_p=NULL;
	int ret,ino;
	printk(KERN_ALERT"MYFS %s is called ",__FUNCTION__);
	ret = myfs_find_entry(parent_dir,(char*)(dentry->d_name.name));
	//Check if the entry already exists in the parent directory ?
	if(ret != -1){
		return -1;
	}
	 ino = get_free_inode_number(ms);
	//See the ms->inode_bmp and get the next inode number
	if(ino == -1){
		printk(KERN_INFO"MYFS(%s) :  Got the free inode number as -1 ",__FUNCTION__);
		return -1;
	}	
		printk(KERN_INFO"MYFS(%s) :  Got the free inode number as %d ",__FUNCTION__,ino);

	//Allocate this new inode 
	//Allocates a new inode Using alloc_inode function of super block adds it to inode list -- 	
	new_inode_p=new_inode(sb);
	if(!new_inode_p)
	{
		printk(KERN_INFO"MYFS(%s) Could Not allocate inode ",__FUNCTION__);
		return -1;
	}
	
	bh =  msi->myfs_super_bh;
	ms->inode_bmp = setbit(ms->inode_bmp,ino);
	new_inode_p->i_ino = ino;
	parent_dir->i_nlink++;
	new_inode_p->i_sb = sb;
	new_inode_p->i_atime = new_inode_p->i_mtime = new_inode_p->i_ctime = CURRENT_TIME_SEC;
	new_inode_p->i_uid = parent_dir->i_uid;
	new_inode_p->i_gid = parent_dir->i_gid;
	

	if(S_ISDIR(mode)){
	printk(KERN_INFO"\nMYFS(%s) Mode is %d and its a directory ",__FUNCTION__,mode);
		new_inode_p->i_mode = MYFS_DIR;
		new_inode_p->i_op = &myfs_dir_inode_ops;   	
		new_inode_p->i_fop = &myfs_dir_ops;
	}	
	else{
	printk(KERN_INFO"\nMYFS(%s) Mode is %d its a file",__FUNCTION__,mode);
		new_inode_p->i_mode = MYFS_FILE;
		new_inode_p->i_op = &myfs_file_inode_ops;   	
		new_inode_p->i_fop = &myfs_file_ops;
	}
	new_inode_p->i_mapping->a_ops = &myfs_aops;

	new_inoi_p = (struct myfs_inode_info*) MYFS_INOI(new_inode_p);
	 if(!new_inoi_p)
        {
                printk(KERN_ALERT"MYFS[%s] INODE info object doesnt exist for new inode",__FUNCTION__);
                return -1;
        }
                printk(KERN_ALERT"MYFS[%s] INODE info object SUCCESS exist for new inode",__FUNCTION__);
	
	new_inoi_p->ino = ino;
	
	new_inoi_p->blockno = ino+1;
	new_inoi_p->filetype =new_inode_p->i_mode;
	new_inoi_p->uid = new_inode_p->i_uid;
	new_inoi_p->gid = new_inode_p->i_gid;
	new_inoi_p->size = 0;
	new_inoi_p->atime = new_inoi_p->mtime = new_inoi_p->ctime = CURRENT_TIME_SEC;
	new_inoi_p->magic = 0xdeadbeef;
	/*REMEMBER IMPORTANT */ 
 	sb->s_dirt=1;
	insert_inode_hash(new_inode_p);
	d_instantiate(dentry,new_inode_p);
	/*
	 *FILL IN INODE INFORMAION FOR A DENTRY  	
	 * PARAMS -- DENTRY TO COMPLETE
	 * INODE TO ATTACH TO THIS DENTRY
	 *
	 * THIS TURNS NEGATIVE DENTRY -- INTO PRODUCTIVE FULL MEMBERS
	 * THIS ASSUMES THAT THE INODE COUNT HAS BEEN INCREMENTED BY THE CALLER 
	 * TO INDOCATE THAT IT IS NOW IN USE BY THE DCACHE
	 * */
		
	//Add the information as a dir entry to parent inode	
	if(myfs_dir_add(parent_dir,ino,dentry->d_name.name )){
	printk(KERN_INFO"MYFS(%s) Could Not add to parent directory ",__FUNCTION__)	;
	return -1;
	}
	mark_buffer_dirty(bh);
	mark_inode_dirty(new_inode_p);
	mark_inode_dirty(parent_dir);
	return 0;
}//function myfs_create

struct dentry * myfs_lookup(struct inode* parent_inode,struct dentry * child_dentry,struct nameidata* nd )
{
	int ino;
	struct inode * reqd_inode;

	printk(KERN_ALERT"MYFS: Trying to Lookup the dentry name %s ",child_dentry->d_name.name);			
	printk(KERN_ALERT"MYFS: superblock is at [0x%p] ",parent_inode->i_sb);		
	/*GET THE INODE NUMBER FOR THAT NAME IN PARENT INODE*/
	ino = myfs_find_entry(parent_inode,(char *)(child_dentry->d_name.name));

	if(ino<=0)
	{
	printk(KERN_ALERT"MYFS [%s] NAME[%s] Lookup Failed",__FUNCTION__,child_dentry->d_name.name);
	reqd_inode= NULL;		
	}
	else
	{
		printk(KERN_ALERT"myfs_iget called from lookup function -- for sb %p and inode number %d",parent_inode->i_sb,ino);
		/*GET INODE FOR THIS INODE NUMBER OBTAINED */
		reqd_inode= myfs_iget(parent_inode->i_sb,ino);
		if(!reqd_inode){
			printk(KERN_ALERT"MYFS: READ inode is null for inodenumber [%d]",ino);
			return ERR_PTR(-EINVAL);
		}
		/*ATTACH TH EINODE WTH GIVEN DENTRY AND RETURN DENTRY */
	}//else
	d_add(child_dentry,reqd_inode);
	return NULL;	
}

static int  delete_dir_entry(struct inode * parent_inode,int ret)
{
	int blockno = parent_inode->i_ino +1;
	struct buffer_head* bh = NULL;
	struct super_block * sb = NULL;
	struct myfs_dir_entry * mde = NULL;
	struct myfs_dir_entry * prev_mde = NULL;
	printk(KERN_ALERT"i m here 11");	
	sb = parent_inode->i_sb ; 

	bh = sb_bread(sb ,blockno) ;
	if(!bh){
		printk(KERN_ALERT"MYFS[%s] Could Not READ data block number %d of parent inode",__FUNCTION__,blockno);
		return -1;
	}

	mde = (struct myfs_dir_entry* )bh->b_data;
	prev_mde = mde;
	printk(KERN_ALERT"i m here 22");	

	while(mde->ino > 0)
	{
		if(mde->ino == ret){
			printk(KERN_ALERT"i m here 33");	
				
			prev_mde->rec_len = prev_mde->rec_len + mde->rec_len;			  				
			break;		 //IMPORTANT
		}
		else{
		prev_mde= mde;
		mde = (struct myfs_dir_entry* )((char* )mde + mde->rec_len);	
		
		}
	}
	printk(KERN_ALERT"i m here 44");	
	mark_buffer_dirty(bh);
	brelse(bh);
	printk(KERN_ALERT"i m here 55");	
	return 0;
}

int myfs_unlink(struct inode* parent_inode,struct dentry* dentry)
{
//	struct inode* child_inode  = dentry->d_inode;
	struct buffer_head * bh = NULL;
	int ret ; 
	struct super_block * sb = NULL;
	struct myfs_sb_info * msi = NULL;
	struct myfs_super_block * ms = NULL;

	ret = myfs_find_entry(parent_inode,(char*)(dentry->d_name.name));
	if(ret <0){
		printk(KERN_ALERT"MYFS : (%s) There is no entry named %s in inode nnumber %lud",__FUNCTION__,dentry->d_name.name,parent_inode->i_ino);
	return -1;
	}
	if(delete_dir_entry(parent_inode,ret) <0)
	{
		printk(KERN_ALERT"MYFS : (%s) Error in delete directory entry in parent inode",__FUNCTION__);
		return -1;
	}
	sb = parent_inode->i_sb;
	msi=  MYFS_SBI(sb);
	ms = msi->ms;
	bh = msi->myfs_super_bh;
  	ms->inode_bmp = clearbit(ms->inode_bmp,ret);
	mark_buffer_dirty(bh);	
	inode_dec_link_count(dentry->d_inode);
	return 0 ;
	/*RERMEMBER WE ARE REMOVING INODE BY SAYING THAT IN BITMAP THAT POSOTION IS ZERO THAT INODE NUMBER CAN 
 * 	BE REUSED NOW NEXT TIME ...
 * 	I DONT UNDERSTAND NOW BY DECREMENTING LINK COUNT USING FUNCTION INODE_DEC_LINK_COUNT	
 * 	i GUESS ITS UNLINK ONLY ACTUALLY INODE SHDNT BE DELETED IN THIS CASE -- THAT IS EXTRA WE ARE PRESUMING ...*/




}
