/*This can be mounted */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include <linux/parser.h>

#include "myfs.h"
#define DEBUG
#define MOD_AUTHOR "NIDHI MITTAL HADA"
#define MODULE_DESC "A Simple Disk Based fileSystem"



//extern struct kmem_cache *myfs_inode_info_p;
struct kmem_cache *myfs_inode_info_p;

extern struct file_operations myfs_dir_ops;
extern struct file_operations myfs_file_ops;

extern struct inode_operations myfs_dir_inode_ops;
extern struct inode_operations myfs_file_inode_ops ;

extern struct super_operations myfs_super_ops;
extern void myfs_write_super(struct super_block *sb);
extern struct address_space_operations myfs_aops;



extern struct dentry * myfs_lookup(struct inode *inode,struct dentry *dentry, struct nameidata *nd);
extern int myfs_create (struct inode * dir, struct dentry * dentry, int mode, struct nameidata *nd);
extern int myfs_permission(struct inode *inode, int mask);
extern int myfs_readdir(struct file *filp, void *dirent, filldir_t);

int myfs_get_super(struct file_system_type *, int,const char *, void *, struct vfsmount *);
void myfs_kill_super(struct super_block * sb);
static int myfs_fill_super(struct super_block * sb, void * a, int b);
void dump_super_block(struct myfs_super_block *ms);


struct inode * myfs_iget(struct super_block *sb,unsigned long ino);
static int  myfs_get_inode(struct inode* parent_inode,char* name);
static int init_myfs_inodecache(void);
static int myfs_init(void);
static void myfs_exit(void);
static inline int  get_opt(int val , int opt);

int myfs_dir_add(struct inode* parent_dir,unsigned int ino,const char* name);
unsigned32 get_free_inode_number(struct myfs_super_block* ms);


static struct file_system_type myfs = {
	.name = "myfs7",
	.owner = THIS_MODULE,
	.get_sb = myfs_get_super,
	.kill_sb = kill_block_super,
	.fs_flags = FS_REQUIRES_DEV
};

enum {
Opt_Force,Opt_Err 
};

static match_table_t token_table = {
//associates an enumerated integer wiyth a text pattern
// its an array of structure of type match_token
//it should have last element as NULL for being used in parse options function
	{Opt_Force,"force"},
	{Opt_Err,NULL},
};	

static inline int  get_opt(int val , int opt)
{
	return (val) & (opt);
}
static inline void set_opt(int *val, int opt)
{
	*val = *val | opt;
	return;
}

static int parse_options(char* options, struct myfs_sb_info *msi)
{
	char *p;
	substring_t args[MAX_OPT_ARGS]; 
	int token;

	if(!options)
	return 1;
	while((p=strsep(&options,","))  != NULL)
	{
	  if(!*p)
	  continue; 	
	  token = match_token(p,token_table,args);
	switch (token) {
		case Opt_Force:
		set_opt(&(msi->mount_opt),MYFS_MNT_FORCE);
		break;
		default:
		printk(KERN_ALERT"MYFS : Unknown mount option - %s\n ",p);
		return 0;
	 }
	}
	return 1;
}//parse_options




int myfs_dir_add(struct inode* parent_dir,unsigned int ino,const char* name)
{
	struct buffer_head* bh;
	struct super_block * sb = parent_dir->i_sb;
	int blocknum = parent_dir->i_ino+1;
	struct myfs_dir_entry mde;
	struct myfs_dir_entry* mde_p;
	
	bh = sb_bread(sb,blocknum);
	if(!bh){
	printk(KERN_INFO"MYFS(%s) Unable to read blocknum %d",__FUNCTION__,blocknum);	
	return 1;
	}
	
	memset(&mde,0,sizeof(mde));
	mde.ino = ino;
	strcpy(mde.name,name);
	mde.name_len=strlen(name);
	mde.rec_len = sizeof(struct myfs_dir_entry);

	mde_p = (struct myfs_dir_entry*)bh->b_data;
	while(mde_p->ino >0 )
	{
	mde_p = (struct myfs_dir_entry*)((char*)mde_p +mde_p->rec_len); 
	}			
	memcpy(mde_p,&mde,mde.rec_len);
	mark_buffer_dirty(bh);
	brelse(bh);
	return 0;

}

int myfs_find_entry(struct inode* parent_dir, char* name)
{
	struct super_block *sb = parent_dir->i_sb;
	int blocknum = parent_dir->i_ino +1;
	struct buffer_head* bh = NULL;
	unsigned int blocksize = sb->s_blocksize;
	struct myfs_dir_entry *mde= NULL;
	bh = sb_bread(sb,blocknum)	;
	if(!bh){
		printk(KERN_INFO"MYFS(%s) Could Not Obtain Block Number %d",__FUNCTION__,blocknum);
		return 0;
	}
	
	mde=(struct myfs_dir_entry*)bh->b_data;
	
	while(mde->ino > 0 && mde < (struct myfs_dir_entry*)(bh->b_data + blocksize)){
			printk(KERN_INFO"Comparing [%s] with entry [%s]",name,mde->name);
	 	if(strcmp(mde->name,name)==0){
			brelse(bh);
			return mde->ino;	
		}
		mde =(struct myfs_dir_entry*)((char*)mde+mde->rec_len);
	}
	brelse(bh);
	return -1;
}
unsigned32 get_free_inode_number(struct myfs_super_block* ms)
{
 	unsigned32 i,bitmap = ms->inode_bmp;
	printk(KERN_INFO"bitmap is [%u] ",bitmap);
	for(i=0;i<=((sizeof(bitmap)*8)-1);i++)
	{
		if(!(bitmap & (1<<i))){
			return i+1;
		}	
		
	}	
	return -1;
}



static int  myfs_get_inode(struct inode* parent_inode,char* name)
{
	struct super_block * sb ;
	int block_no ;
	struct buffer_head* bh;
	struct myfs_dir_entry * tmp_dirent;
	int count_dirent,tmp_count;
	struct myfs_dir_entry final_dirent;

	printk(KERN_ALERT "MYFS: myfs_get_inode called");
	printk(KERN_ALERT "MYFS: Reading data from block number %d for name:inode %s:%lud",parent_inode->i_blocks,name,parent_inode->i_ino);

	sb = parent_inode->i_sb;
        block_no  = parent_inode->i_blocks;
        if(!(bh = sb_bread(sb,block_no))){ 
	 printk(KERN_ALERT"MYFS: block number %d could not be read",block_no);
	 return 0;
	}
	tmp_dirent = (struct myfs_dir_entry *)bh->b_data;	
	count_dirent = (sb->s_blocksize)/sizeof(struct myfs_dir_entry);	
	tmp_count=1;
	while(tmp_count <= count_dirent )
	{
	  printk(KERN_ALERT"MYFS: READ DIRECTORY ENTRY NAME UNMATCHED %s",tmp_dirent->name);	
	if(strcmp(tmp_dirent->name,name)==0)
	{
	final_dirent.ino = tmp_dirent->ino;
	memcpy(final_dirent.name,tmp_dirent->name,MAX_NAME_LEN);
	final_dirent.rec_len = tmp_dirent->rec_len;
	final_dirent.name_len = tmp_dirent->name_len;
	brelse(bh);
	 return final_dirent.ino; 		
	}	
	tmp_dirent = tmp_dirent + sizeof(struct myfs_dir_entry);
	tmp_count = tmp_count+1;
	}	
	printk(KERN_ALERT"MYFS: NO DIR Entry matched the required name %s RETURNING Inode NUmber = -1",name);

	return -1;
}


struct inode * myfs_iget(struct super_block *sb,unsigned long ino)
{
	struct inode * inode ;
	struct buffer_head *bh;
	struct myfs_inode  *myfs_ino;

	
	inode = iget_locked(sb,ino);
	if(!inode){
	printk(KERN_ALERT"MYFS:Could not get inode by iget_locked fn");
	return ERR_PTR(-ENOMEM);;
	}
	printk(KERN_ALERT"MYFS: inode obtained for inode number %lud ",ino);
	printk(KERN_ALERT "reading inode from disk sb->s_bdev = %p , blocksize = %ld blocknumbr is %d \n ", sb->s_bdev,sb->s_blocksize,MYFS_INODE_BLOCK);
	if(!(bh=(struct buffer_head *)sb_bread(sb,MYFS_INODE_BLOCK))){
		printk(KERN_ALERT"Cannot read inode block %d for inod enumber %ld\n",MYFS_INODE_BLOCK,ino);
		goto free;
	}
	//myfs_ino=(struct myfs_inode *)((char *)(bh->b_data)+ ((ino) * sizeof(struct myfs_inode)));
	//REMEMBER FROM HERE INODE NUMBER 1 IS THE FIRST TNODE STORED IN INODE BLOCK 
	myfs_ino=(struct myfs_inode *)((char *)(bh->b_data)+ ((ino-1) * sizeof(struct myfs_inode)));

	
	inode->i_private = myfs_ino;
	inode->i_mode = myfs_ino->filetype;
	inode->i_size = myfs_ino->size;
	inode->i_uid=myfs_ino->uid;
	inode->i_gid = myfs_ino->gid;
	inode->i_atime = myfs_ino->atime;
	inode->i_ctime = myfs_ino->ctime;
	inode->i_mtime = myfs_ino->mtime;

	inode->i_ino = myfs_ino->ino; // may not be needed 
	//inode->i_blocks = myfs_ino->block_no;
	//inode->i_nlink = myfs_ino->links;
	if(S_TYPE(myfs_ino->filetype,S_IFREG)){
		printk(KERN_INFO "MYFS (%s) : Inode number [%d] is a file\n",__FUNCTION__,myfs_ino->ino);	
		inode->i_fop = &myfs_file_ops;
		inode->i_op = &myfs_file_inode_ops;
	}else if(S_TYPE(myfs_ino->filetype,S_IFDIR)){
		 printk(KERN_INFO "MYFS (%s) : Inode number [%d] is a directory\n",__FUNCTION__,myfs_ino->ino);
		inode->i_op=&myfs_dir_inode_ops; 
		inode->i_fop=&myfs_dir_ops;	
	}else{
	printk(KERN_INFO "MYFS (%s) : Inode number [%lud] is a unknown file\n",__FUNCTION__,ino);
	}
	unlock_new_inode(inode);
 	brelse(bh);
	return inode;	
	free:
	return ERR_PTR(-1);
}//myfs_iget


static int myfs_fill_super(struct super_block *sb, void *data, int silent)
{
	
	struct myfs_sb_info *msi;
	struct myfs_super_block *ms;

	struct inode *root_inode;
	struct buffer_head *bh;
	//struct myfs_inode  *myfs_ino;
	
	msi = kmalloc(sizeof(struct myfs_sb_info),GFP_KERNEL);
	if(!msi)
	return -ENOMEM;
	sb->s_fs_info = msi;
	sb->s_blocksize=MYFS_DEFAULT_BS;
	memset(msi,0,sizeof(*msi));	
	/*Read the super block*/
	printk(KERN_ALERT "sb->s_bdev = %p , blocksize = %ld \n", sb->s_bdev,sb->s_blocksize );
	
	if(!sb_set_blocksize(sb,MYFS_DEFAULT_BS)){
	#ifdef DEBUG
         printk(KERN_INFO "MYFS (%s) : Bad blocksize \n",__FUNCTION__);
	#endif
	return -EACCES;
	}	
	if(!(bh=(struct buffer_head *)sb_bread(sb,MYFS_SUPER_BLOCK)))
	{
	  printk(KERN_ALERT"Cannot read superblock of MYFS \n");	
	   goto free;
	}
	else
	printk(KERN_ALERT"Read superblock of MYFS in block [%d] \n",MYFS_SUPER_BLOCK);
	
	ms =(struct myfs_super_block*)(bh->b_data) ;
	msi->ms =ms;
	msi->mount_opt=0;
	if(!parse_options((char*)data,msi)){
		kfree(msi);
		return -EINVAL;
	}
	
	if(!get_opt(msi->mount_opt , MYFS_MNT_FORCE)  && (ms->dirty != MYFS_CLEAN) ){
#ifdef DEBUG
        printk(KERN_CRIT "MYFS (%s) : File system not clean 	due to previous dirty unmount.Use \"-o force\" option to force mount. \n",__FUNCTION__);
#endif
        kfree(msi);
	return -EINVAL;
	}
	
	printk(KERN_ALERT"MYFS:%s I am marking the sb dirty %d\n",__FUNCTION__,msi->ms->dirty);
	//msi->ms->dirty = MYFS_DIRTY;
	msi->ms->dirty = 0x01;
	printk(KERN_ALERT"MYFS:%s I have marked the sb dirty %d\n",__FUNCTION__,msi->ms->dirty);
	msi->myfs_super_bh = bh;
	dump_super_block(ms);
	
	sb->s_magic = ms->s_magic;
	printk(KERN_ALERT"MYFS: Got the Magic Number [%lud]",sb->s_magic);
	if(sb->s_magic != MYFS_SUPER_MAGIC){
		printk(KERN_ALERT"Not a Myfs FIlesystem");
		goto free;
	}
	sb->s_op = &myfs_super_ops;
		

	/*------------READ THE INODE ------------*/
	root_inode=myfs_iget(sb,MYFS_ROOT_INO);
	printk(KERN_ALERT"MYFS: root inode is read by myfs_iget ");

	/*give inode number to already fully initialized super block -- to get root inode*/

	sb->s_root = d_alloc_root(root_inode);
	if(!sb->s_root){
		iput(root_inode);
		printk(KERN_ALERT"MYFS: Cannot read root inode");	
		goto free;
	}	
	 	printk(KERN_ALERT"MYFS: dentry is attached  ");
	if(!S_ISDIR(root_inode->i_mode)){
		dput(sb->s_root);
		sb->s_root = NULL;
		printk(KERN_ALERT"MYFS:Error reading root inode");
		goto free;
	}	 
	root_inode->i_fop = &myfs_dir_ops;
	root_inode->i_op = &myfs_dir_inode_ops;
	myfs_write_super(sb);
			
	 	printk(KERN_ALERT"MYFS: root inode's mode is checked successfully  ");
	return 0;
	free:
	 	kfree(msi);	
		return -EINVAL;	 		
}


void dump_super_block(struct myfs_super_block *ms)
{
printk(KERN_ALERT"dumping super block of MYFS : root %d,block_bmp = %d,inode_bmp =%d,block_size = %d,blocks = %d,s_magic = 0x%x dirty bit  = %d \n",ms->root_inode,ms->block_bmp,ms->inode_bmp,ms->block_size,ms->total_blocks,ms->s_magic,ms->dirty);
}



int myfs_get_super(struct file_system_type *fs, int flags,const char *dev_name , void * data , struct vfsmount *mnt)
{
#ifdef DEBUG
        printk(KERN_INFO "MYFS (%s) : called\n",__FUNCTION__);
#endif
 return get_sb_bdev(fs,flags,dev_name,data,myfs_fill_super,mnt);
}

static void myfs_exit(void)
{
#ifdef DEBUG
	printk(KERN_INFO"Unregistering Filesystem myfs");
#endif
	unregister_filesystem(&myfs);
	kmem_cache_destroy(myfs_inode_info_p);
	return;
}


static  void init_once(void * foo)
{

	struct myfs_inode_info* mi=( struct myfs_inode_info *)foo;
/*
 * These are initializations that only need to be done
 * once, because the fields are idempotent across use
 * of the inode, so let the slab aware of that.
 */
	
	inode_init_once(&(mi->vino));
	return;

}


static int init_myfs_inodecache(void)
{
	myfs_inode_info_p = kmem_cache_create("myfs_inode_info_cache",sizeof(struct myfs_inode_info),0,(SLAB_RECLAIM_ACCOUNT|SLAB_MEM_SPREAD),init_once);

	if(!myfs_inode_info_p){
		printk(KERN_ALERT"MYFS(%s) Could Not Create cache ",__FUNCTION__);
		return -ENOMEM;
	}
	return 0;
}


static int myfs_init(void)
{
	int err;
	#ifdef DEBUG
	printk(KERN_INFO"Registering Filesystem myfs \n");
	#endif
	err = register_filesystem(&myfs);
	err = init_myfs_inodecache();
	return err;
}
module_init(myfs_init);
module_exit(myfs_exit);

//MODULE_AUTHOR(MOD_AUTHOR);
//MODULE_LICENSE("GPL");
//MODULE_DESCRIPTION(MODULE_DESC);

