#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include <linux/types.h>
#include <linux/statfs.h>

#include "myfs.h"

extern struct kmem_cache *myfs_inode_info_p;

struct inode * myfs_alloc_inode(struct super_block *sb);
static void myfs_destroy_inode(struct inode *inode);


int myfs_write_inode(struct inode *inode, int wait);

void myfs_put_super(struct super_block *sb);
void myfs_write_super(struct super_block *sb);
static int myfs_statfs (struct dentry * dentry, struct kstatfs * buf);


struct super_operations myfs_super_ops={
	.alloc_inode = myfs_alloc_inode,
	.destroy_inode = myfs_destroy_inode,
	.write_inode = myfs_write_inode,
	.put_super = myfs_put_super,
	.write_super = myfs_write_super	,	
	.statfs = myfs_statfs,

};

static int myfs_count_blocks(struct myfs_super_block *ms)
{
	int i ;

	for(i=0;i<=((sizeof(ms->inode_bmp)*8) -1);i++)
	{
	 	if(!((ms->inode_bmp) & (1<<i))){
		return i ;	
		}
	}
	return i;
}


static int myfs_statfs (struct dentry * dentry, struct kstatfs *buf)
{
	struct super_block *sb = dentry->d_sb;
	struct myfs_sb_info *msi = MYFS_SBI(sb);
	struct myfs_super_block *ms = msi->ms;



	buf->f_type =  (long)MYFS_SUPER_MAGIC;
	buf->f_bsize = (long)sb->s_blocksize;
	buf->f_namelen = (long)MAX_NAME_LEN;
	buf ->f_blocks = buf->f_files = sizeof(ms->inode_bmp)*8;
	buf->f_ffree = buf->f_bfree = buf->f_bavail = (buf->f_blocks - myfs_count_blocks(ms));	
return 0 ;
}




struct inode * myfs_alloc_inode(struct super_block *sb){
	struct myfs_inode_info* msi=kmem_cache_alloc(myfs_inode_info_p,GFP_KERNEL);
	#ifdef DEBUG
	printk(KERN_ALERT"MYFS(%s) : Called",__FUNCTION);
	#endif
	if(!msi){
		printk(KERN_ALERT"MYFS(%s) Error in allocation of new inode returning null",__FUNCTION__);
		return NULL;
	}
	msi->vino.i_version=1;
	return(&(msi->vino));
}

static void myfs_destroy_inode(struct inode *inode)
{
	#ifdef DEBUG
printk(KERN_INFO "MYFS (%s) : Called for inode number inode->i_ino[%lu] and inode_info->ino [%d]\n",__FUNCTION__,inode->i_ino,NNGFS_I(inode)->ino);
#endif
        kmem_cache_free(myfs_inode_info_p,MYFS_INOI(inode));
}



int myfs_write_inode(struct inode *inode, int wait){
	struct super_block * sb = inode->i_sb;
	int ino = inode->i_ino;
	struct buffer_head *bh = NULL;
	struct myfs_inode  mi ;
	
	#ifdef DEBUG
        printk(KERN_INFO "MYSS (%s) : Called  for inode [%lu]\n",__FUNCTION__,inode->i_ino);
	#endif	
	memset(&mi,0,sizeof(struct myfs_inode));
	
	mi.ino = ino;
	mi.block_no = inode->i_ino+1;
	mi.filetype = inode->i_mode;
	mi.uid = inode->i_uid;
	mi.gid = inode->i_gid;
	mi.size = inode->i_size;
	mi.atime = inode->i_atime;
	mi.ctime = inode->i_ctime;
	mi.mtime = inode->i_mtime;
	mi.magic = MYFS_SUPER_MAGIC;

	bh = sb_bread(sb,MYFS_INODE_BLOCK);
	if(!bh){
		printk(KERN_ALERT"MYFS (%s) Not able to read the inode block for inode number %lud",__FUNCTION__,inode->i_ino);
		return -1;
	}		
	
	memcpy(  ((struct myfs_inode*)(bh->b_data+((inode->i_ino -1)*sizeof(struct myfs_inode)))) , &mi, sizeof(struct myfs_inode));	
	mark_buffer_dirty(bh);
	brelse(bh);
	return 0 ;
}


void myfs_put_super(struct super_block *sb){
	struct myfs_sb_info *msi= sb->s_fs_info;
	struct buffer_head* sbh =msi->myfs_super_bh;  
	brelse(sbh);
	kfree(msi);
	return;		
}
void myfs_write_super(struct super_block *sb){
	struct myfs_sb_info *msi= sb->s_fs_info;
        struct buffer_head* sbh =msi->myfs_super_bh;
	printk(KERN_ALERT"MYFS[%s] is called ",__FUNCTION__);
	mark_buffer_dirty(sbh);
	sb->s_dirt=0;
	return;
}
